from typing import Callable, Dict, Optional, Tuple, Union

import pytorch_lightning as pl
import torch
from torch import nn
from torch.nn import functional as F
from torch_geometric.nn import GCNConv, GINConv, global_mean_pool
from tqdm import tqdm


class VGGBlock(nn.Module):
    """A VGG-style convolutional block consisting of a
    convolution layer, batch normalization, and ReLU activation.

    Args:
        in_channels: (int) Number of input channels.
        out_channels: (int) Number of output channels.
        kernel_size: (int, optional) Size of the convolution kernel. Default is 3.
        padding: (int, optional) Padding for the convolution layer. Default is 1.
        stride: (int, optional) Stride for the convolution layer. Default is 1.

    Returns:
        Output tensor after applying the convolution, batch normalization, and ReLU activation.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        padding: int = 1,
        stride: int = 1,
    ):
        super(VGGBlock, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size, stride, padding)
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, images):
        x = self.conv(images)
        x = self.bn(x)
        x = self.relu(x)
        return x


class ResnetBlock(nn.Module):
    """A ResNet-style residual block consisting of two convolution layers with skip connections.

    Args:
        in_channels (int): Number of input channels.
        out_channels (int): Number of output channels.
        kernel_size (int, optional): Size of the convolution kernel. Default is 3.
        padding (int, optional): Padding for the convolution layers. Default is 1.
        stride (int, optional): Stride for the convolution layers. Default is 1.

    Returns:
        Output tensor after applying residual connection and ReLU activation.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int = 3,
        padding: int = 1,
        stride: int = 1,
    ):
        super(ResnetBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.bn1 = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            out_channels, out_channels, kernel_size=kernel_size, padding=padding
        )
        self.bn2 = nn.BatchNorm2d(out_channels)

    def forward(self, images):
        identity = images
        out = self.conv1(images)
        out = self.bn1(out)
        out = self.relu(out)
        out = self.conv2(out)
        out = self.bn2(out)
        out += identity
        out = self.relu(out)

        return out


class InceptionBlock(nn.Module):
    """An Inception-style block consisting of multiple
    convolution branches operating on different scales.

    Args:
        in_channels: (int) Number of input channels.
        ch1x1: (int) Number of output channels for the 1x1 convolution branch.
        ch3x3_reduce: (int) Number of output channels for the 1x1 convolution before the 3x3 convolution.
        ch3x3: (int) Number of output channels for the 3x3 convolution branch.
        ch5x5_reduce: (int) Number of output channels for the 1x1 convolution before the 5x5 convolution.
        ch5x5: (int) Number of output channels for the 5x5 convolution branch.
        pool_proj: (int) Number of output channels for the 1x1 convolution after the max pooling branch.

    Returns:
        Concatenated output tensor from all branches.
    """

    def __init__(
        self,
        in_channels: int,
        ch1x1: int,
        ch3x3_reduce: int,
        ch3x3: int,
        ch5x5_reduce: int,
        ch5x5: int,
        pool_proj: int,
    ):
        super(InceptionBlock, self).__init__()

        self.branch1x1 = nn.Conv2d(in_channels, ch1x1, kernel_size=1)
        self.branch3x3 = nn.Sequential(
            nn.Conv2d(in_channels, ch3x3_reduce, kernel_size=1),
            nn.Conv2d(ch3x3_reduce, ch3x3, kernel_size=3, padding=1),
        )
        self.branch5x5 = nn.Sequential(
            nn.Conv2d(in_channels, ch5x5_reduce, kernel_size=1),
            nn.Conv2d(ch5x5_reduce, ch5x5, kernel_size=5, padding=2),
        )
        self.branch_pool = nn.Sequential(
            nn.MaxPool2d(kernel_size=3, stride=1, padding=1),
            nn.Conv2d(in_channels, pool_proj, kernel_size=1),
        )

    def forward(self, images):
        branch1x1 = self.branch1x1(images)
        branch3x3 = self.branch3x3(images)
        branch5x5 = self.branch5x5(images)
        branch_pool = self.branch_pool(images)
        outputs = [branch1x1, branch3x3, branch5x5, branch_pool]
        return torch.cat(outputs, 1)


class DSCBlock(nn.Module):
    """Depthwise separable convolution, which consists of a depthwise
    convolution followed by a pointwise convolution.

    Args:
        in_channels: (int) Number of input channels.
        out_channels: (int) Number of output channels.
        kernel_size: (int, optional) Size of the convolution kernel. Default is 3.
        stride: (int, optional) Stride for the convolution layers. Default is 1.
        padding: (int, optional) Padding for the convolution layers. Default is 1.

    Returns:
        Output tensor after applying depthwise and pointwise convolutions, batch normalization, and ReLU activation.
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, stride=1, padding=1):
        super(DSCBlock, self).__init__()
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            groups=in_channels,
            bias=False,
        )
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, images):
        x = self.depthwise(images)
        x = self.pointwise(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


_blocks = {
    "VGG_block": VGGBlock,
    "Resnet_block": ResnetBlock,
    "Inception_block": InceptionBlock,
    "DSC_block": DSCBlock,
}


def create_simple_CNN(
    config: Dict[int, Dict[str, Union[str, Dict]]],
    in_channels: int,
    shape: Tuple[int, int] = None,
):
    """Creates a simple CNN based on the provided configuration.

    Args:
        config: (Dict) Configuration dictionary where each key represents a layer/block and its corresponding parameters.
        in_channels: (int) Number of input channels for the first layer.
        shape: (Tuple[int, int], optional) Input shape for the CNN. If provided, checks that the final output shape is valid.

    Returns: (nn.Sequential, Tuple[int, int]) or nn.Sequential
        cnn: (nn.Sequential)
            Sequential CNN model.
        shape: (Tuple[int, int], optional)
            Final output shape, if provided.
    """

    modules = list()

    for idx, block_config in tqdm(config.items()):
        block_type = block_config["type"]
        params = block_config.get("params", {})

        if block_type != "Inception_block":
            module = _blocks[block_type](in_channels=in_channels, **params)
            in_channels = params.get("out_channels", in_channels)
            if shape is not None:
                kernel_size = params.get("kernel_size", 3)
                padding = params.get("padding", 1)
                stride = params.get("stride", 1)
                shape = (
                    (shape[0] - kernel_size + 2 * padding) // stride + 1,
                    (shape[1] - kernel_size + 2 * padding) // stride + 1,
                )
                if shape[0] < 1 or shape[1] < 1:
                    raise ValueError(
                        """Your CNN backbone is too large for the input shape! 
                    Increase resolution or consider reducing the number of layers/removing stride/adding padding."""
                    )
        else:
            module = _blocks[block_type](in_channels=in_channels, **params)
            in_channels = sum(
                [params.get(key) for key in ["ch1x1", "ch3x3", "ch5x5", "pool_proj"]]
            )

        modules.append(module)

    cnn = nn.Sequential(*modules)

    if shape is not None:
        return cnn, shape
    return cnn


class SimpleRNN(nn.Module):
    def __init__(
        self,
        rnn_type,
        input_size,
        hidden_size,
        num_layers=1,
        bidirectional=False,
        pre_rnn_linear_size=None,
    ):
        """A simple recurrent neural network (RNN) module
        that supports RNN, LSTM, and GRU architectures.

        Args:
            rnn_type: (str) Type of RNN ('RNN', 'LSTM', or 'GRU').
            input_size: (int) Number of input features.
            hidden_size: (int) Number of hidden units.
            num_layers: (int, optional) Number of RNN layers. Default is 1.
            bidirectional: (bool, optional) Whether the RNN is bidirectional. Default is False.
            pre_rnn_linear_size: (int, optional) Size of the optional linear layer before the RNN.

        Returns:
            Output tensor after the RNN and fully connected layer.
        """
        super(SimpleRNN, self).__init__()
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bidirectional = bidirectional

        # Optional linear layer before the RNN
        if pre_rnn_linear_size is not None:
            self.pre_rnn_linear = nn.Linear(input_size, pre_rnn_linear_size)
            self.input_size = pre_rnn_linear_size  # Update input_size for the RNN layer
        else:
            self.pre_rnn_linear = None
            self.input_size = input_size  # Original input size

        # Set the RNN layer based on the rnn_type parameter
        if rnn_type == "RNN":
            self.rnn = nn.RNN(
                self.input_size,
                hidden_size,
                num_layers,
                batch_first=True,
                bidirectional=bidirectional,
            )
        elif rnn_type == "LSTM":
            self.rnn = nn.LSTM(
                self.input_size,
                hidden_size,
                num_layers,
                batch_first=True,
                bidirectional=bidirectional,
            )
        elif rnn_type == "GRU":
            self.rnn = nn.GRU(
                self.input_size,
                hidden_size,
                num_layers,
                batch_first=True,
                bidirectional=bidirectional,
            )
        else:
            raise ValueError(
                f"Unsupported rnn_type: {rnn_type}. Choose from 'RNN', 'LSTM', 'GRU'."
            )

    def forward(self, sequences, lengths):
        # Apply the optional linear layer before the RNN
        x = sequences
        if self.pre_rnn_linear is not None:
            x = self.pre_rnn_linear(x)

        # Pack the padded sequence
        packed_x = nn.utils.rnn.pack_padded_sequence(
            x, lengths, batch_first=True, enforce_sorted=False
        )

        # Forward pass through the RNN
        packed_out, hidden = self.rnn(packed_x)

        # Unpack the sequence
        # out, _ = nn.utils.rnn.pad_packed_sequence(packed_out, batch_first=True)

        # If the RNN is bidirectional, concatenate the hidden states from both directions
        if self.bidirectional:
            if isinstance(
                hidden, tuple
            ):  # For LSTM, hidden is a tuple (hidden_state, cell_state)
                hidden = torch.cat((hidden[0][-2], hidden[0][-1]), dim=1)
            else:  # For RNN and GRU
                hidden = torch.cat((hidden[-2], hidden[-1]), dim=1)
        else:
            if isinstance(hidden, tuple):  # For LSTM
                hidden = hidden[0][-1]
            else:  # For RNN and GRU
                hidden = hidden[-1]
        # Pass the last hidden state through the fully connected layer
        return hidden


class VitNet(nn.Module):
    """Parent class for a vision-and-text network that
    fuses CNN and RNN-based representations using
    concatenation or addition.

    Args:
        CNN: (nn.Module) CNN backbone for processing image data.
        RNN: (nn.Module) RNN backbone for processing sequence data.
        fusion_mode: (str, optional) Fusion mode ('concat' or 'add'). Default is 'concat'.
        activation: (nn.Module, optional) Activation function applied after fusion. Default is None.
        embed_dim: (int, optional) Embedding dimension for the projected features. Default is 128.
    """

    def __init__(self, CNN, RNN, fusion_mode="concat", activation=None, embed_dim=32):
        super(VitNet, self).__init__()
        self.CNN = CNN
        self.RNN = RNN
        self.fusion_mode = fusion_mode
        self.activation = activation
        self.image_proj = nn.LazyLinear(embed_dim)
        self.rnn_proj = nn.LazyLinear(embed_dim)
        self.flat = nn.Flatten()

    def forward(self, images, sequences, lengths):
        x_1 = self.image_proj(self.flat(self.CNN(images)))

        projected_sequences = self.rnn_projection(sequences)
        packed_input = nn.utils.rnn.pack_padded_sequence(
            projected_sequences, lengths, batch_first=True, enforce_sorted=False
        )
        packed_output, x_2 = self.RNN(packed_input)
        _, _ = nn.utils.rnn.pad_packed_sequence(
            packed_output, batch_first=True, padding_value=0
        )

        if self.fusion_mode == "add":
            x = x_1 + x_2
        elif self.fusion_mode == "concat":
            x = torch.cat([x_1, x_2], dim=1)
        else:
            raise NotImplementedError(
                f"Fusion mode {self.fusion_mode} is not implemented."
            )

        if self.activation is not None:
            x = self.activation(x)

        return x


class VitNetWithCrossAttention(VitNet):
    """Child class extending VitNet by adding cross-attention
    between image and sequence representations.

    Args:
        CNN: (nn.Module) CNN backbone for processing image data.
        RNN: (nn.Module) RNN backbone for processing sequence data.
        input_dim: (int) Input dimension for the RNN.
        projected_dim: (int) Dimension of the projected sequence representation.
        cross_attention_fusion_mode: (str, optional) Fusion mode after cross-attention ('concat' or 'add'). Default is 'concat'.
        activation: (nn.Module, optional) Activation function applied after fusion. Default is None.
        embed_dim: (int, optional) Embedding dimension for the projected features. Default is 128.
        return_attention_weights: (bool, optional) Whether to return attention weights. Default is False.
    """

    def __init__(
        self,
        CNN,
        RNN,
        input_dim: int,
        projected_dim: int,
        cross_attention_fusion_mode="concat",
        activation=None,
        embed_dim=128,
        return_attention_weights=False,
    ):
        super(VitNetWithCrossAttention, self).__init__(
            CNN,
            RNN,
            input_dim,
            projected_dim,
            fusion_mode="concat",
            activation=activation,
            embed_dim=embed_dim,
        )
        self.cross_attention_layer = nn.MultiheadAttention(embed_dim, num_heads=8)
        self.cross_attention_fusion_mode = cross_attention_fusion_mode
        self.return_attention_weights = return_attention_weights
        self.padding_idx = RNN.padding_idx

    def forward(self, images, sequences, lengths):
        # First, run the parent class's forward method to get the basic fusion output
        x = super().forward(images, sequences, lengths)

        # Cross-attention mechanism
        x_1 = self.image_proj(self.CNN(images))
        projected_sequences = self.sequence_projection(sequences)
        packed_input = nn.utils.rnn.pack_padded_sequence(
            projected_sequences, lengths, batch_first=True, enforce_sorted=False
        )
        packed_output, _ = self.RNN(packed_input)
        all_hidden, _ = nn.utils.rnn.pad_packed_sequence(
            packed_output, batch_first=True, padding_value=0
        )

        query = self.image_proj(x_1).unsqueeze(0)  # (1, batch_size, embed_dim)
        key = self.rnn_proj(all_hidden).transpose(0, 1)
        key_padding_mask = sequences[:, :, 0] == self.padding_idx
        context_vector, attention_weights = self.cross_attention_layer(
            query, key, key, key_padding_mask=key_padding_mask
        )

        output = context_vector.squeeze(0)
        if self.cross_attention_fusion_mode == "concat":
            x = torch.cat([x, output], dim=1)
        else:
            x = output

        if self.activation is not None:
            x = self.activation(x)

        if self.return_attention_weights:
            return x, attention_weights

        return x


class GCN(torch.nn.Module):
    """A graph convolutional network (GCN)
    with optional learnable node embeddings.

    Args:
        num_nodes: (int) Number of nodes in the graph.
        feature_dim: (int) Dimensionality of node features.
        embedding_dim: (int) Dimensionality of the learnable node embeddings.
        layer_sizes: (List[int]) List of hidden layer sizes for each GCN layer.
        out_channels: (int) Number of output channels.
        use_embeddings: (bool, optional) Whether to use learnable embeddings. Default is True.

    Returns:
        Output tensor after graph convolutions and global mean pooling.
    """

    def __init__(
        self,
        num_nodes,
        feature_dim,
        embedding_dim,
        layer_sizes,
        out_channels,
        use_embeddings=True,
    ):
        super(GCN, self).__init__()
        self.use_embeddings = use_embeddings
        if use_embeddings:
            self.embeddings = torch.nn.Embedding(
                num_nodes, embedding_dim
            )  # Learnable embeddings

        in_channels = feature_dim + (embedding_dim if use_embeddings else 0)
        self.convs = torch.nn.ModuleList()  # List of GCN layers
        for hidden_channels in layer_sizes:
            self.convs.append(GCNConv(in_channels, hidden_channels))
            in_channels = hidden_channels  # Update in_channels for next layer

    def forward(self, graphs):
        if self.use_embeddings:
            # Combine node features and learnable embeddings
            node_embeddings = self.embeddings.weight[
                graphs.mapping
            ]  # Retrieve embeddings for each node
            x = torch.cat([graphs.x, node_embeddings], dim=1)
        else:
            # Use only node features
            x = graphs.x

        # Apply each GCN layer with optional edge weights
        for conv in self.convs:
            x = conv(x, graphs.edge_index, edge_weight=graphs.edge_attr)
            x = F.relu(x)

        # Global pooling (mean pooling)
        x = global_mean_pool(x, graphs.batch)

        return x


class GIN(torch.nn.Module):
    """A graph isomorphism network (GIN) with shared
    embeddings for nodes across graphs.

    Args:
        num_common_nodes: (int) Number of shared nodes across graphs.
        feature_dim: (int) Dimensionality of node features.
        embedding_dim: (int) Dimensionality of the shared node embeddings.
        layer_sizes: (List[int]) List of hidden layer sizes for each GIN layer.
        out_channels: (int) Number of output channels.

    Returns:
        Output tensor after GIN layers and global mean pooling.
    """

    def __init__(
        self, num_common_nodes, feature_dim, embedding_dim, layer_sizes, out_channels
    ):
        super(GIN, self).__init__()
        # Shared embeddings across graphs
        self.embeddings = torch.nn.Embedding(num_common_nodes, embedding_dim)

        # Determine the input dimension for the first GIN layer
        in_channels = feature_dim + embedding_dim

        self.convs = torch.nn.ModuleList()  # List of GIN layers
        for hidden_channels in layer_sizes:
            nn = nn.Sequential(
                nn.Linear(in_channels, hidden_channels),
                nn.ReLU(),
                nn.Linear(hidden_channels, hidden_channels),
            )
            self.convs.append(GINConv(nn))
            in_channels = hidden_channels  # Update in_channels for next layer

    def forward(self, graphs):
        # Retrieve the shared embeddings using the common index
        shared_embeddings = self.embeddings(graphs.common_index)

        # Concatenate node features with the shared embeddings
        x = torch.cat([graphs.x, shared_embeddings], dim=1)

        # Apply each GIN layer
        for conv in self.convs:
            x = conv(x, graphs.edge_index)
            x = F.relu(x)

        # Global pooling (mean pooling)
        x = global_mean_pool(x, graphs.batch)

        return x


class BaseModel(pl.LightningModule):
    """A base PyTorch Lightning module for neural networks
    with optional custom optimizer and scheduler.

    Args:
        backbone: (Union[nn.ModuleList, nn.Module]) The feature extraction backbone model.
        output_size: (int) Size of the final output layer.
        hidden_layers: (Tuple, optional) Tuple of hidden layer sizes. Default is empty.
        activation: (nn.Module, optional) Activation function to apply between layers. Default is ReLU.
        learning_rate: (float, optional) Learning rate for the optimizer. Default is 1e-3.
        optimizer_class: (Callable, optional) Optimizer class to use. Default is AdamW.
        optimizer_params: (dict, optional) Additional parameters for the optimizer. Default is None.
        scheduler_class: (Callable, optional) Scheduler class to use. Default is None.
        scheduler_params: (dict, optional) Additional parameters for the scheduler. Default is None.
        loss_fn: (Callable, optional) Loss function to use. Default is None.

    Returns:
        Output after the forward pass through the network.
    """

    def __init__(
        self,
        backbone: Union[nn.ModuleList, nn.Module],
        output_size,
        hidden_layers: Tuple = (),
        activation=nn.ReLU(),
        learning_rate: float = 1e-3,
        optimizer_class: Callable = torch.optim.AdamW,
        optimizer_params: Optional[dict] = None,
        scheduler_class: Optional[Callable] = None,
        scheduler_params: Optional[dict] = None,
        loss_fn: Optional[Callable] = None,
    ):
        super().__init__()

        self.backbone = backbone

        modules = list()

        for hidden_units in hidden_layers:
            modules.append(nn.LazyLinear(hidden_units))
            modules.append(activation)
            in_features = hidden_units
        modules.append(nn.LazyLinear(output_size))
        self.head = nn.ModuleList(modules)

        self.loss_fn = loss_fn
        self.LR = learning_rate

        # Optimizer and Scheduler configuration
        self.optimizer_class = optimizer_class
        self.optimizer_params = optimizer_params if optimizer_params is not None else {}
        self.scheduler_class = scheduler_class
        self.scheduler_params = scheduler_params if scheduler_params is not None else {}

        self.flat = nn.Flatten()

    def forward(self, x):
        if isinstance(self.backbone, nn.Sequential):
            x = self.backbone(*x.values())
        else:
            x = self.backbone(**x)
        if len(x.size()) == 4:
            x = self.flat(x)
        for layer in self.head:
            x = layer(x)
        return x

    def configure_optimizers(self):
        optimizer = self.optimizer_class(
            self.parameters(), lr=self.LR, **self.optimizer_params
        )

        if self.scheduler_class is not None:
            scheduler = self.scheduler_class(optimizer, **self.scheduler_params)
            return [optimizer], [scheduler]
        else:
            return optimizer

    def training_step(self, batch, batch_idx):
        y = batch.pop("y")
        out = self(batch)
        loss = self.loss_fn(out, y)
        self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        return loss

    def validation_step(self, batch, batch_idx):
        y = batch.pop("y")
        out = self(batch)
        loss = self.loss_fn(out, y)
        self.log("valid_loss", loss, on_step=False, on_epoch=True, prog_bar=True)


from typing import Callable, Optional, Tuple, Union

import torch
import torch.nn as nn
import torchmetrics


class Classifier(BaseModel):
    """A classification model built on top of the BaseModel,
    with additional accuracy, precision, recall, and F1-score tracking.

    Args:
        backbone: (Union[nn.ModuleList, nn.Module]) The feature extraction backbone model.
        n_classes: (int) Number of output classes.
        classifier_hidden_layers: (Tuple, optional) Tuple of hidden layer sizes for the classifier. Default is empty.
        classifier_activation: (nn.Module, optional) Activation function to use in the classifier. Default is ReLU.
        learning_rate: (float, optional) Learning rate for the optimizer. Default is 1e-3.
        optimizer_class: (Callable, optional) Optimizer class to use. Default is AdamW.
        optimizer_params: (dict, optional) Additional parameters for the optimizer. Default is None.
        scheduler_class: (Callable, optional) Scheduler class to use. Default is None.
        scheduler_params: (dict, optional) Additional parameters for the scheduler. Default is None.

    Returns:
        Output logits after the forward pass through the classifier.
    """

    def __init__(
        self,
        backbone: Union[nn.ModuleList, nn.Module],
        n_classes,
        classifier_hidden_layers=(),
        classifier_activation=nn.ReLU(),
        learning_rate=1e-3,
        optimizer_class: Callable = torch.optim.AdamW,
        optimizer_params: Optional[dict] = None,
        scheduler_class: Optional[Callable] = None,
        scheduler_params: Optional[dict] = None,
    ):
        super().__init__(
            backbone=backbone,
            output_size=n_classes,
            hidden_layers=classifier_hidden_layers,
            activation=classifier_activation,
            learning_rate=learning_rate,
            optimizer_class=optimizer_class,
            optimizer_params=optimizer_params,
            scheduler_class=scheduler_class,
            scheduler_params=scheduler_params,
            loss_fn=nn.CrossEntropyLoss(),
        )

        self.accuracy = torchmetrics.Accuracy(
            task="binary" if n_classes == 2 else "multiclass", num_classes=n_classes
        )

        self.precision = torchmetrics.Precision(
            task="binary" if n_classes == 2 else "multiclass",
            num_classes=n_classes,
            average=None,  # Per-class precision
        )

        self.recall = torchmetrics.Recall(
            task="binary" if n_classes == 2 else "multiclass",
            num_classes=n_classes,
            average=None,  # Per-class recall
        )

        self.f1 = torchmetrics.F1Score(
            task="binary" if n_classes == 2 else "multiclass",
            num_classes=n_classes,
            average=None,  # Per-class F1-score
        )

        # For macro average metrics
        self.macro_precision = torchmetrics.Precision(
            task="binary" if n_classes == 2 else "multiclass",
            num_classes=n_classes,
            average="macro",  # Macro-average precision
        )

        self.macro_recall = torchmetrics.Recall(
            task="binary" if n_classes == 2 else "multiclass",
            num_classes=n_classes,
            average="macro",  # Macro-average recall
        )

        self.macro_f1 = torchmetrics.F1Score(
            task="binary" if n_classes == 2 else "multiclass",
            num_classes=n_classes,
            average="macro",  # Macro-average F1-score
        )

        self.prob = nn.Softmax(dim=1)

    def validation_step(self, batch, batch_idx):
        y = batch.pop("y")  # Assuming that "y" is the ground truth labels
        out = self(batch)  # Forward pass
        loss = self.loss_fn(out, y)
        out = self.prob(out)
        logits = torch.argmax(out, dim=1)

        # Calculate metrics
        accu = self.accuracy(logits, y)
        precision = self.precision(logits, y)
        recall = self.recall(logits, y)
        f1 = self.f1(logits, y)

        macro_precision = self.macro_precision(logits, y)
        macro_recall = self.macro_recall(logits, y)
        macro_f1 = self.macro_f1(logits, y)

        # Log loss and accuracy
        self.log("valid_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        self.log("val_acc_step", accu, on_step=False, on_epoch=True, prog_bar=False)

        # Log precision, recall, and F1-score for each class
        for i in range(len(precision)):
            self.log(
                f"val_precision_class_{i}",
                precision[i],
                on_step=False,
                on_epoch=True,
                prog_bar=False,
            )
            self.log(
                f"val_recall_class_{i}",
                recall[i],
                on_step=False,
                on_epoch=True,
                prog_bar=False,
            )
            self.log(
                f"val_f1_class_{i}", f1[i], on_step=False, on_epoch=True, prog_bar=False
            )

        # Log macro averages
        self.log(
            "val_macro_precision",
            macro_precision,
            on_step=False,
            on_epoch=True,
            prog_bar=False,
        )
        self.log(
            "val_macro_recall",
            macro_recall,
            on_step=False,
            on_epoch=True,
            prog_bar=False,
        )
        self.log("val_macro_f1", macro_f1, on_step=False, on_epoch=True, prog_bar=False)


class Regressor(BaseModel):
    """A regression model built on top of the BaseModel,
    using mean squared error loss.

    Args:
        backbone: (Union[nn.ModuleList, nn.Module]) The feature extraction backbone model.
        output_dim: (int) Dimensionality of the regression output.
        regressor_hidden_layers: (Tuple, optional) Tuple of hidden layer sizes for the regressor. Default is empty.
        regressor_activation: (nn.Module, optional) Activation function to use in the regressor. Default is ReLU.
        learning_rate: (float, optional) Learning rate for the optimizer. Default is 1e-3.
        optimizer_class: (Callable, optional) Optimizer class to use. Default is AdamW.
        optimizer_params: (dict, optional) Additional parameters for the optimizer. Default is None.
        scheduler_class: (Callable, optional) Scheduler class to use. Default is None.
        scheduler_params: (dict, optional) Additional parameters for the scheduler. Default is None.

    Returns:
        Regression output after the forward pass.
    """

    def __init__(
        self,
        backbone: Union[nn.ModuleList, nn.Module],
        output_dim,
        regressor_hidden_layers=(),
        regressor_activation=nn.ReLU(),
        learning_rate=1e-3,
        optimizer_class: Callable = torch.optim.AdamW,
        optimizer_params: Optional[dict] = None,
        scheduler_class: Optional[Callable] = None,
        scheduler_params: Optional[dict] = None,
    ):
        super().__init__(
            backbone=backbone,
            output_size=output_dim,
            hidden_layers=regressor_hidden_layers,
            activation=regressor_activation,
            learning_rate=learning_rate,
            optimizer_class=optimizer_class,
            optimizer_params=optimizer_params,
            scheduler_class=scheduler_class,
            scheduler_params=scheduler_params,
            loss_fn=nn.MSELoss(),
        )



from typing import List, Tuple

import numpy as np
import pandas as pd
from scipy import stats
from scipy.signal import find_peaks, savgol_filter


# =========================== BLINKS DETECTION ===========================
# Helper function
def _interpolate_nans(
    array: np.ndarray, timestamps: np.ndarray, gap_dur: float = np.inf
) -> np.ndarray:
    """Function finds sequences of NaN values, selects ones with
    duration <= 'gap_dur' and linearly interpolates them.

    Args:
        array: array with NaNs.
        timestamps: timestamps of array.
        gap_dur: threshold gap duration.

    Returns:
        array with interpolated gaps.
    """
    assert len(array.shape) == 1, "Only 1D array is allowed."
    assert (
        array.shape == timestamps.shape
    ), "'array' and 'timestamps' must correspond in shape."
    assert gap_dur > 0, "Gap duration must be positive."

    # Find index for nans where gaps are longer than 'gap_dur' samples
    mask = np.isnan(array)

    # If there are no nans, return
    if not np.any(mask):
        return array

    # Find onsets and offsets of gaps
    d = np.diff(np.concatenate((np.array([0]), mask * 1, np.array([0]))))
    onsets = np.where(d == 1)[0]
    offsets = np.where(d == -1)[0]

    # Decrease offsets come too late by -1
    if np.any(offsets >= len(array)):
        idx = np.where(offsets >= len(array))[0][0]
        offsets[idx] = offsets[idx] - 1

    dur = timestamps[offsets] - timestamps[onsets]

    # If the gaps are longer than 'gaps', replace temporarily with other values
    for i, on in enumerate(onsets):
        if dur[i] > gap_dur:
            array[onsets[i] : offsets[i]] = -1000

    # New is-nan mask after 'gaps' grasp
    mask = np.isnan(array)
    imask = ~mask
    array[mask] = np.interp(mask.nonzero()[0], imask.nonzero()[0], array[imask])

    # Put nans back
    array[array == -1000] = np.nan

    return array


# Helper function
def _nearest_odd_integer(x: float) -> int:
    """Method computes nearest odd integer to 'x'."""
    return int(2 * np.floor(np.abs(x) / 2) + 1) * np.sign(x)


# Helper function
def _mask2bounds(mask: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Method constructs onset and offset arrays of indices based on mask.
    Onset - starting index of segment (segment is continuous block of ones),
    offset - ending index of blink.

    Args:
        mask: boolean array.

    Returns:
        [onsets, offsets].
    """

    # Find segments
    d = np.diff(np.hstack((0, mask, 0)))
    onsets = np.where(d == 1)[0]
    offsets = np.where(d == -1)[0] - 1

    # Match onsets with offsets
    if len(offsets) > len(onsets):
        if onsets[0] > offsets[0]:
            offsets = offsets[1:]
        else:
            offsets = offsets[:-1]
    elif len(offsets) < len(onsets):
        if onsets[0] > offsets[0]:
            onsets = onsets[1:]
        else:
            onsets = onsets[:-1]

    return onsets, offsets


# Helper function
def _merge_blinks(
    blink_onsets: List | np.ndarray,
    blink_offsets: List | np.ndarray,
    min_dur: int,
    min_separation: int,
    blink_properties: List | np.ndarray = None,
) -> List[List]:
    """Method merges blinks given onsets and offsets, also collapses too short ones.

    Args:
        blink_onsets: onsets of blinks, ms.
        blink_offsets: offsets of blinks, ms.
        min_dur: min duration of blink, ms.
        min_separation: min duration between blinks, ms.
        blink_properties: array of lists, properties of blinks.

    Returns:
        array of triples (onset, offset, duration).
    """
    have_properties = blink_properties is not None

    # Merge blink candidate close together, and remove short, isolated ones
    new_onsets = []
    new_offsets = []
    new_properties = []
    change_onset = True
    temp_onset = None

    for i, onset in enumerate(blink_onsets):
        if change_onset:
            temp_onset = blink_onsets[i]

        if i < len(blink_onsets) - 1:
            if (blink_onsets[i + 1] - blink_offsets[i]) < min_separation:
                change_onset = False
            else:
                change_onset = True

                # Remove blink with too short duration
                if (blink_offsets[i] - temp_onset) < min_dur:
                    continue

                new_onsets.append(temp_onset)
                new_offsets.append(blink_offsets[i])
                if have_properties:
                    new_properties.append(blink_properties[i, :])
        else:
            # Remove blink with too short duration
            if (blink_offsets[i] - temp_onset) < min_dur:
                continue

            new_onsets.append(temp_onset)
            new_offsets.append(blink_offsets[i])
            if have_properties:
                new_properties.append(blink_properties[i, :])

    # Compute durations and convert to array
    blinks = []
    for i in range(len(new_onsets)):
        dur = new_offsets[i] - new_onsets[i]
        blink_data = [new_onsets[i], new_offsets[i], dur]
        if have_properties:
            blink_data += list(new_properties[i])
        blinks.append(blink_data)

    return blinks


# Helper function
def _indices_to_values(
    onsets: np.ndarray, offsets: np.ndarray, timestamps: np.ndarray, Fs: int = None
) -> Tuple[np.ndarray, np.ndarray]:
    """Method converts index-based onsets/offsets to
    timestamp-based onsets/offsets.

    Args:
        onsets: indexes of starting blink indexes.
        offsets: indexes of ending blink indexes.
        timestamps: data recording timestamps.
        Fs: sample rate of eye tracker, Hz.

    Returns:
        onsets, offsets.
    """

    # Convert onsets/offsets to ms
    blinks = []
    for onset, offset in zip(onsets, offsets):
        blinks.append([timestamps[onset], timestamps[offset]])

    if Fs is not None:
        assert Fs > 0
        # Remove blinks with on-, or offsets that happened in a period of missing data
        # (i.e., where samples are completely lost, for some reason)
        idx = np.where(
            np.diff(timestamps) > (2 * 1 / Fs * 1000)
        )  # Missing data where deltaT > 2 * 1/Fs

        for i, blink in enumerate(blinks):
            for idx_k in idx[0]:
                if np.logical_and(
                    blink[0] >= timestamps[idx_k], blink[0] <= timestamps[idx_k + 1]
                ) or np.logical_and(
                    blink[1] >= timestamps[idx_k], blink[1] <= timestamps[idx_k + 1]
                ):
                    blinks.pop(i)
                    break

    onsets = np.array([b[0] for b in blinks])
    offsets = np.array([b[1] for b in blinks])

    return onsets, offsets


# Helper function
def _apply_moving_average(
    pupil_signal: np.ndarray,
    timestamps: np.ndarray,
    is_na: np.ndarray,
    max_window_size: float,
) -> np.ndarray:
    """Method applies moving average for pupillometry data. Nan values are remained.

    Args:
        pupil_signal: sizes of pupil.
        timestamps: data recording timestamps.
        is_na: boolean array representing whether pupil is nan or not.
        max_window_size: maximum size of smoothing window, in milliseconds.

    Returns:
        smoothed pupil sizes.
    """

    if len(pupil_signal) <= 2:
        return pupil_signal
    assert len(pupil_signal) == len(timestamps) == len(is_na)
    assert max_window_size > 0.0

    n = len(pupil_signal)
    tmp = pupil_signal.copy()
    tmp = np.where(is_na, 0, tmp)
    csum_size = np.cumsum(tmp)
    csum_time = timestamps - np.roll(timestamps, shift=1)
    csum_time[0] = 0
    csum_time = np.cumsum(csum_time)

    smoothed_sizes = np.zeros((n,))
    smoothed_sizes[0] = pupil_signal[0]

    window_end = -1
    for i in range(1, n):
        if is_na[i]:
            window_end = -1
            smoothed_sizes[i] = np.nan
            continue

        window_end = max([window_end - 1, i + 1])
        while (
            window_end < n
            and not is_na[window_end]
            and (csum_time[window_end] - csum_time[i - 1]) <= max_window_size
        ):
            window_end += 1
        smoothed_sizes[i] = (csum_size[window_end - 1] - csum_size[i - 1]) / (
            window_end - i
        )

    return smoothed_sizes


# Blinks Detector
def detect_blinks_pupil_missing(
    pupil_signal: np.ndarray,
    timestamps: np.ndarray,
    min_separation: int = 100,
    min_dur: int = 20,
    smoothing_window_size: int = 10,  # 10ms
    return_mask: bool = False,
) -> pd.DataFrame | Tuple[pd.DataFrame, np.ndarray]:
    """Method detects blinks based on size of pupil and
    missing recordings (NaN) in its data. Result is boolean array of the same
    length as pupil_signal, with 1 indicating blink, 0 - not blink.

    Args:
        pupil_signal: size of right or left pupil.
        timestamps: data recording timestamps.
        smoothing_window_size: maximum size of smoothing window, ms.
        min_separation: min time interval between detected blinks, ms.
        min_dur: min duration of blink, ms.
        return_mask: bool = False

    Returns:
        detected blinks dataframe.
    """

    assert not np.isnan(timestamps).any(), "Timestamps must not be nan."
    assert len(pupil_signal) > 10, "There must be at least 10 recordings."

    # get mask for NaN values
    is_na = np.isnan(pupil_signal).astype(np.int32)
    assert np.sum(is_na) > 0, (
        "This algorithm is based on missing values in pupil size recordings data, but"
        "provided 'pupil_signal' array does not contain NaNs."
    )

    # smooth using moving average filter
    smoothed_pupil_signal = _apply_moving_average(
        pupil_signal, timestamps, is_na, smoothing_window_size
    )

    # calculate difference between neighbor values
    diff = smoothed_pupil_signal - np.roll(smoothed_pupil_signal, shift=1)
    diff = np.where(np.roll(is_na, shift=1), smoothed_pupil_signal, diff)
    diff[0] = 0

    # naive approach - missing value is blink
    is_blink = is_na
    n = len(pupil_signal)

    # improve approach - expand blink boundaries using monotonic sequences
    for i in range(1, n - 1):
        if is_blink[i]:
            continue

        # monotonic pupil size decrease
        if is_blink[i + 1]:
            left = i
            assert not np.isnan(diff[left])
            while left >= 0 and not is_blink[left] and diff[left] < 0:
                is_blink[left] = 1
                left -= 1

        # monotonic pupil size increase
        if is_blink[i - 1]:
            right = i
            while right < n and not is_blink[right] and diff[right] > 0:
                is_blink[right] = 1
                right += 1

    onsets, offsets = _mask2bounds(is_blink)
    onsets, offsets = _indices_to_values(onsets, offsets, timestamps)

    # Merge blinks closer than x ms, and remove short blinks
    blinks = _merge_blinks(onsets, offsets, min_dur, min_separation)
    df = pd.DataFrame(blinks, columns=["onset", "offset", "duration"])
    if return_mask:
        return df, is_blink
    return df


# Blinks Detector
def detect_blinks_pupil_vt(
    pupil_signal: np.ndarray,
    timestamps: np.ndarray,
    Fs: int,
    gap_dur: int = 20,
    min_dur: int = 20,
    remove_outliers: bool = False,
    window_len: int = None,
    min_pupil_size: int = 2,
    min_separation: int = 50,
    return_mask: bool = False,
) -> pd.DataFrame | Tuple[pd.DataFrame, np.ndarray]:
    """Method detects blinks based on pupil sizes and change of
    pupil sizes.

    Args:
        pupil_signal: size of right of left pupil, mm.
        timestamps: data recording timestamps, ms.
        Fs: sample rate of eye tracker, Hz.
        gap_dur: max gaps between periods of data loss, ms.
        min_dur: min duration of blink, ms.
        remove_outliers: whether to remove outliers.
        window_len: size of window to use if 'remove_outliers' = True.
        min_pupil_size: min value of pupil size considered to be recorded correctly, mm.
        min_separation: min time interval between detected blinks, ms.
        return_mask: if True, then mask showing blink classification is also returned.

    Returns:
        detected blinks dataframe.
    """

    # Remove outliers
    if remove_outliers:
        if np.isnan(window_len):
            window_len_samples = len(pupil_signal)
        else:
            window_len_samples = int(
                Fs / 1000 * window_len
            )  # in ms window over which to exclude outliers

        ps = pupil_signal.copy()
        ps[ps < min_pupil_size] = np.nan

        for k in np.arange(len(ps) - window_len_samples + 1):
            temp = pupil_signal[k : (k + window_len_samples)].copy()

            if len(temp) == 0:
                continue

            m = np.nanmean(temp)
            sd = np.nanstd(temp)
            idx = (temp > (m + 3 * sd)) | (temp < (m - 3 * sd))
            temp[idx] = np.nan
            ps[k : (k + window_len_samples)] = temp

        pupil_signal = ps

    # Interpolate short periods of data loss
    pupil_signal = _interpolate_nans(pupil_signal, timestamps, gap_dur=gap_dur)

    # Convert to bounds and clean up
    is_blink = np.isnan(pupil_signal) * 1
    onsets, offsets = _mask2bounds(is_blink)
    onsets, offsets = _indices_to_values(onsets, offsets, timestamps, Fs=Fs)

    # Merge blinks closer than x ms, and remove short blinks
    blinks = _merge_blinks(onsets, offsets, min_dur, min_separation)
    df = pd.DataFrame(blinks, columns=["onset", "offset", "duration"])
    if return_mask:
        return df, is_blink
    return df


# Blinks Detector
def detect_blinks_eo(
    eye_openness_signal: np.ndarray,
    timestamps: np.ndarray,
    Fs: int,
    gap_dur: int = 30,
    filter_length: int = 25,
    min_blink_length: int = 15,
    min_amplitude: int = 0.1,
    min_separation: int = 100,
    return_eo_vel: bool = False,
) -> pd.DataFrame | Tuple[pd.DataFrame, np.ndarray]:
    """Method detects blinks based on Eye Openness (EO) signal.

    Args:
        eye_openness_signal: eye openness signal.
        timestamps: data recording timestamps, ms.
        Fs: sample rate of eye tracker, Hz.
        gap_dur: max gaps between periods of data loss, ms.
        filter_length: length of Savitzky-Golay filter, ms.
        min_blink_length: min length of detected blinks, ms.
        min_separation: min time interval between detected blinks, ms.
        min_amplitude: fraction of fully opened eye.
        return_eo_vel: if True, then computed velocity of EO signal is returned.

    Returns:
        detected blinks dataframe.
    """

    ms_to_sample = Fs / 1000
    sample_to_ms = 1000 / Fs

    # Assumes the eye is mostly open during the trial
    fully_open = np.nanmedian(eye_openness_signal, axis=0)
    min_amplitude = fully_open * min_amplitude  # Equivalent to height in 'find_peaks'

    # detection parameters in samples
    distance_between_blinks = 1
    min_blink_length = min_blink_length * ms_to_sample
    filter_length = _nearest_odd_integer(filter_length * ms_to_sample)

    # Interpolate gaps
    eye_openness_signal = _interpolate_nans(
        eye_openness_signal, timestamps, gap_dur=int(gap_dur)
    )

    # Filter eyelid signal and compute
    eye_openness_signal_filtered = savgol_filter(
        eye_openness_signal, filter_length, 2, mode="nearest"
    )
    eye_openness_signal_vel = (
        savgol_filter(eye_openness_signal, filter_length, 2, deriv=1, mode="nearest")
        * Fs
    )

    # Velocity threshold for on-, and offsets
    T_vel = stats.median_abs_deviation(eye_openness_signal_vel, nan_policy="omit") * 3

    # Turn blink signal into something that looks more like a saccade signal
    eye_openness_signal_inverse = (
        eye_openness_signal_filtered - np.nanmax(eye_openness_signal_filtered)
    ) * -1
    peaks, properties = find_peaks(
        eye_openness_signal_inverse,
        height=None,
        distance=distance_between_blinks,
        width=min_blink_length,
    )

    # Filter out not so 'prominent peaks'
    """
    The prominence of a peak may be defined as the least drop in height
     necessary in order to get from the summit [peak] to any higher terrain.
    """
    idx = properties["prominences"] > min_amplitude
    peaks = peaks[idx]
    for key in properties.keys():
        properties[key] = properties[key][idx]

    # Find peak opening/closing velocity by searching for max values
    # within a window from the peak
    blink_properties = []
    for i, peak_idx in enumerate(peaks):
        # Width of peak
        width = properties["widths"][i]

        ### Compute opening/closing velocity
        # First eye opening velocity (when eyelid opens after a blink)
        peak_right_idx = np.nanargmax(
            eye_openness_signal_vel[peak_idx : int(peak_idx + width)]
        )
        peak_right_idx = np.nanmin([peak_right_idx, len(eye_openness_signal_vel)])
        idx_max_opening_vel = int(peak_idx + peak_right_idx)
        time_max_opening_vel = timestamps[idx_max_opening_vel]
        opening_velocity = np.nanmax(
            eye_openness_signal_vel[peak_idx : int(peak_idx + width)]
        )

        # Then eye closing velocity (when eyelid closes in the beginning of a blink)
        peak_left_idx = (
            width
            - np.nanargmin(
                eye_openness_signal_vel[np.max([0, int(peak_idx - width)]) : peak_idx]
            )
            + 1
        )
        peak_left_idx = np.nanmax([peak_left_idx, 0])
        idx_max_closing_vel = int(peak_idx - peak_left_idx + 1)
        time_max_closing_vel = timestamps[idx_max_closing_vel]
        closing_velocity = np.nanmin(
            eye_openness_signal_vel[np.max([0, int(peak_idx - width)]) : peak_idx]
        )

        # Identify on and offsets (go from peak velocity backward/forward)
        temp = eye_openness_signal_vel[idx_max_opening_vel:]
        if np.any(temp <= (T_vel / 3)):
            offset = np.where(temp <= (T_vel / 3))[0][0]
        else:
            offset = len(temp)

        # make sure the blink period stop when encountering nan-data
        # If it does, make the opening phase parameters invalid
        if np.any(np.isnan(temp)):
            offset_nan = np.where(np.isnan(temp))[0][0]
            offset = np.min([offset, offset_nan])

        offset_idx = int(idx_max_opening_vel + offset - 1)

        temp = np.flip(eye_openness_signal_vel[:idx_max_closing_vel])
        if np.any(temp >= -T_vel):
            onset = np.where(temp >= -T_vel)[0][0]
        else:
            onset = 0

        if np.any(np.isnan(temp)):
            onset_nan = np.where(np.isnan(temp))[0][0]
            onset = np.min([onset, onset_nan])

        onset_idx = int(idx_max_closing_vel - onset)

        # Compute openness at onset, peak, and offset
        openness_at_onset = eye_openness_signal_filtered[onset_idx]
        openness_at_offset = eye_openness_signal_filtered[offset_idx]
        openness_at_peak = eye_openness_signal_filtered[peak_idx]

        # Compute amplitudes for closing and opening phases
        closing_amplitude = np.abs(openness_at_onset - openness_at_peak)
        opening_amplitude = np.abs(openness_at_offset - openness_at_peak)

        distance_onset_peak_vel = np.abs(
            eye_openness_signal_filtered[onset_idx]
            - eye_openness_signal_filtered[idx_max_closing_vel]
        )  # mm
        timediff_onset_peak_vel = (
            np.abs(onset_idx - idx_max_closing_vel) * sample_to_ms
        )  # ms

        # Onset and peak cannot be too close in space and time
        if (distance_onset_peak_vel < 0.1) or (timediff_onset_peak_vel < 10):
            continue

        if np.min([opening_velocity, np.abs(closing_velocity)]) < (T_vel * 2):
            continue

        blink_properties.append(
            [
                timestamps[onset_idx],
                timestamps[offset_idx],
                timestamps[offset_idx] - timestamps[onset_idx],
                timestamps[peak_idx],
                openness_at_onset,
                openness_at_offset,
                openness_at_peak,
                time_max_opening_vel,
                time_max_closing_vel,
                opening_velocity,
                closing_velocity,
                opening_amplitude,
                closing_amplitude,
            ]
        )

    # Are there any blinks found?
    if len(blink_properties) == 0:
        bp = []
    else:
        # Merge blinks too close together in time
        blink_temp = np.array(blink_properties)
        blink_onsets = blink_temp[:, 0]
        blink_offsets = blink_temp[:, 1]

        bp = _merge_blinks(
            blink_onsets,
            blink_offsets,
            int(min_blink_length),
            min_separation,
            blink_properties=blink_temp[:, 3:],
        )

    # Convert to dataframe
    df = pd.DataFrame(
        bp,
        columns=[
            "onset",
            "offset",
            "duration",
            "time_peak",
            "openness_at_onset",
            "openness_at_offset",
            "openness_at_peak",
            "time_peak_opening_velocity",
            "time_peak_closing_velocity",
            "peak_opening_velocity",
            "peak_closing_velocity",
            "opening_amplitude",
            "closing_amplitude",
        ],
    )

    idx = df.index[df["openness_at_peak"] < 0]
    df.loc[idx, "openness_at_peak"] = 0
    if return_eo_vel:
        return df, eye_openness_signal_vel
    return df



import warnings
from abc import abstractmethod
from typing import Any, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from numpy.typing import NDArray

from eyefeatures.features.extractor import BaseTransformer
from eyefeatures.utils import (Types, _calc_dt, _select_regressions,
                               _split_dataframe)


class StatsTransformer(BaseTransformer):
    """Base class for statistical features. Aggregate function strings must be
    compatible with `pandas`. Expected dataframe with fixations.

    Args:
        features_stats: Dictionary of format {'feature_1': ['statistic_1', 'statistic_2'], ...}.
        x: X coordinate column name.
        y: Y coordinate column name.
        t: timestamp column name.
        duration: duration column name (milliseconds expected).
        dispersion: fixation dispersion column name.
        aoi: Area Of Interest column name(-s). If provided, features can be calculated inside
            the specified AOI.
        calc_without_aoi: if True, then, in addition to AOI-wise features, calculate regular features
            ignoring AOI.
        pk: primary key.
        return_df: whether to return output as DataFrame or numpy array.
        warn: whether to enable warnings.
    """

    def __init__(
        self,
        features_stats: Dict[str, List[str]],
        x: str = None,
        y: str = None,
        t: str = None,
        duration: str = None,  # TODO consider units, i.e. ps, ns, ms.
        dispersion: str = None,
        aoi: str | List[str] = None,
        calc_without_aoi: bool = False,
        pk: List[str] = None,
        return_df: bool = True,
        warn: bool = True,
    ):
        super().__init__(
            x=x,
            y=y,
            t=t,
            duration=duration,
            dispersion=dispersion,
            aoi=aoi,
            pk=pk,
            return_df=return_df,
        )
        self.features_stats = features_stats
        # feature -- i.e. saccade length/speed
        self.feature_names_in = list(features_stats.keys())
        self.available_feats = None
        self.eps = 1e-20
        self.aoi = aoi
        self.calc_without_aoi = calc_without_aoi
        self.aoi_mapper = None

        self.warn = warn

        self.feature_names_in_ = None

    @staticmethod
    def _err_no_col(f, c):
        return f"Requested feature {f} requires {c} for calculation."

    def _check_feature_names(self, X, *, reset):
        """"""
        # Since feature names must always be provided with statistics to
        # calculate and are restricted to certain set of available features,
        # this method is irrelevant
        raise AttributeError("Use '_check_features_stats' instead.")

    def _validate_data(
        self,
        X="no_validation",
        y="no_validation",
        reset=True,
        validate_separately=False,
        cast_to_ndarray=True,
        **check_params,
    ):
        """"""
        # Same reason as for _check_feature_names
        raise AttributeError("This method must not be used.")

    @abstractmethod
    def _check_params(self):
        """
        Method checks that all requested features could be calculated with provided data.
        """
        ...

    def _check_features_stats(self):
        """
        Method checks `self.features_stats` for correct feature names (i.e. keys).
        """
        err_msg = (
            lambda f: f"Feature '{f}' is not supported. Must be one of: "
            f"{', '.join(self.available_feats)}."
        )
        for feat in self.feature_names_in:
            assert feat in self.available_feats, err_msg(feat)

        self._check_params()

    # method called on fit
    def _check_aoi_fit(self, X):
        if self.aoi is not None:  # check if aoi columns contain any NaNs
            assert isinstance(self.aoi, str) or isinstance(
                self.aoi, list
            ), f"`aoi` must be str or List[str], got {type(self.aoi)}."

            if isinstance(self.aoi, str):
                self.aoi = [self.aoi]

            assert (
                "" not in self.aoi
            ), 'Empty string "" as value in `aoi` columns is not allowed.'

            for aoi_col in self.aoi:
                if aoi_col != "":
                    aoi_view = X[aoi_col]
                    if aoi_view.isnull().values.any():
                        raise RuntimeError(
                            f"Passed column '{aoi_col}' for AOI contains NaNs."
                        )

        self._preprocess_aoi(X)

    def _preprocess_aoi(self, X: pd.DataFrame):
        if self.aoi is not None:
            self.aoi_mapper = dict()

            if self.calc_without_aoi:
                self.aoi_mapper[""] = [""]

            for aoi_col in self.aoi:
                aoi_view = X[aoi_col]
                self.aoi_mapper[aoi_col] = aoi_view.drop_duplicates().values.tolist()

        else:
            self.aoi_mapper = {"": [""]}  # convenience placeholder

    # method called on transform
    def _check_aoi_transform(self, X: pd.DataFrame):
        if self.aoi is not None:  # check if aoi column contains any NaNs
            assert (
                "" not in self.aoi
            ), 'Empty string "" as value in `aoi` columns is not allowed.'

            for aoi_col in self.aoi:
                if aoi_col == "":  # lib placeholder
                    continue

                aoi_view = X[aoi_col]
                if aoi_view.isnull().values.any():
                    raise RuntimeError(
                        f"Passed column '{aoi_col}' for AOI contains NaNs."
                    )

                for v in aoi_view:
                    assert (
                        v in self.aoi_mapper[aoi_col]
                    ), f"Unknown AOI value {v} was not seen during `fit` in '{aoi_col}'."

    @property
    @abstractmethod
    def _fp(self) -> str:
        """
        Feature prefix to use in feature names.
        """
        ...

    @abstractmethod
    def _calc_feats(
        self, X: pd.DataFrame, features: List[str], transition_mask: NDArray
    ) -> List[Tuple[str, pd.Series]]:
        """
        Method calculates features passed to constructor, i.e. keys of `self.features_stats`.
        In case of `SaccadeFeatures`, it returns dictionary `{'length': np.array, 'velocity': np.array, ...}`.
        `transition_mask` is boolean mask of the same shape as X, i-th value is False if X's i-th value is
        first fixation in block. Block is defined as sequential fixations in same AOI, maximum by inclusion
        (which means that block cannot contain another block). Thus, each AOI is split in blocks and
        first fixation in each block is then removed.
        """
        ...

    def _calc_with_aoi(
        self, feat_nms: List[str], X: pd.DataFrame, aoi_col: str, aoi_val: Any
    ) -> List[Tuple[str, pd.Series]]:
        """
        Helper function to calculate features based on `aoi_col` (name for aoi column)
        and `aoi_val` (aoi value in this column).
        """
        if aoi_col == "":  # internal lib placeholder
            transition_mask = np.ones(len(X)).astype(bool)
            feats: List[Tuple[str, pd.Series]] = self._calc_feats(
                X, feat_nms, transition_mask
            )

        else:
            X_aoi = X[X[aoi_col] == aoi_val]
            all_aoi = X[aoi_col]
            all_transition_mask: pd.Series = all_aoi == all_aoi.shift(1)
            transition_mask = all_transition_mask[all_aoi == aoi_val].values

            feats: List[Tuple[str, pd.Series]] = self._calc_feats(
                X_aoi, feat_nms, transition_mask
            )

        return feats

    def fit(self, X: pd.DataFrame, y=None):
        self._check_features_stats()
        self._check_aoi_fit(X)

        # all output features that will appear on transform
        self.feature_names_in_ = []
        for aoi_col in self.aoi_mapper:
            for aoi_val in self.aoi_mapper[aoi_col]:
                for feat_nm in self.features_stats:
                    for stat in self.features_stats[feat_nm]:
                        self.feature_names_in_.append(
                            f"{self._fp}_{feat_nm}_{aoi_col}[{aoi_val}]_{stat}"
                            if aoi_col != ""
                            else f"{self._fp}_{feat_nm}_{stat}"
                        )
        return self

    def transform(self, X: pd.DataFrame) -> Union[pd.DataFrame, NDArray]:
        if self.features_stats is None:
            return X if self.return_df else X.values

        self._check_aoi_transform(X)

        feat_nms = list(self.features_stats.keys())
        gathered_stats = []
        column_nms = []

        if self.pk is None:
            groups: Types.EncodedPartition = [("0", X)]
        else:
            groups: Types.EncodedPartition = _split_dataframe(
                X, self.pk
            )  # split by unique groups

        group_ids = []
        for group_id, group_X in groups:
            group_ids.append(group_id)
            gathered_stats_group = []

            for aoi_col in self.aoi_mapper:
                for aoi_val in self.aoi_mapper[aoi_col]:
                    group_feats: List[Tuple[str, pd.Series]] = self._calc_with_aoi(
                        feat_nms, group_X, aoi_col, aoi_val
                    )

                    add_cols_nms = len(group_ids) == 1
                    for feat_nm, feat_arr in group_feats:
                        feat_stats: List[str] = self.features_stats[feat_nm]

                        if not feat_arr.empty:  # group_X with AOI was not empty
                            stats_group = [feat_arr.apply(stat) for stat in feat_stats]

                        else:  # no AOI for given group
                            stats_group = [None for _ in feat_stats]

                        gathered_stats_group.extend(stats_group)
                        # TODO remove, have self.features_names_in_ on fit. But order matters,
                        #  so not removed for now
                        if add_cols_nms:
                            column_nms.extend(
                                [
                                    (
                                        f"{self._fp}_{feat_nm}_{aoi_col}[{aoi_val}]_{stat}"
                                        if aoi_col != ""
                                        else f"{self._fp}_{feat_nm}_{stat}"
                                    )
                                    for stat in feat_stats
                                ]
                            )

            gathered_stats.append(gathered_stats_group)

        assert len(self.feature_names_in_) == len(column_nms)
        for i in range(len(column_nms)):
            assert (
                self.feature_names_in_[i] == column_nms[i]
            ), f"Fit: {self.feature_names_in_}\nTransform: {column_nms}."
        stats_df = pd.DataFrame(
            data=gathered_stats, columns=column_nms, index=group_ids
        )

        return stats_df if self.return_df else stats_df.values


class SaccadeFeatures(StatsTransformer):
    """Saccade Features Transformer.
    The transformer identifies saccades from fixations and extract
    their features.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.available_feats = ("length", "acceleration", "speed")

    @property
    def _fp(self) -> str:
        return "sac"

    def _check_params(self):
        for feat in self.feature_names_in:
            assert self.x is not None, self._err_no_col(feat, "x")
            assert self.y is not None, self._err_no_col(feat, "y")
            if feat in ("speed", "acceleration"):
                assert self.t is not None, self._err_no_col(feat, "t")

    def _calc_feats(
        self, X: pd.DataFrame, features: List[str], transition_mask: NDArray
    ) -> List[Tuple[str, pd.Series]]:
        feats = []

        dx: pd.Series = X[self.x].diff()
        dy: pd.Series = X[self.y].diff()
        dr = np.sqrt(dx**2 + dy**2)
        dt = (
            _calc_dt(X, self.duration, self.t)
            if any(map(lambda f: f != "length", features))
            else None
        )

        for feat_nm in features:
            if feat_nm == "length":
                sac_len = dr
                feat_arr = sac_len[transition_mask]
            elif feat_nm == "acceleration":
                # Acceleration: dx = v0 * t + 1/2 * a * t^2.
                # Above formula is law of uniformly accelerated motion TODO consider direction
                sac_acc: pd.DataFrame = dr / (dt**2 + self.eps) * 1 / 2
                feat_arr = sac_acc[transition_mask]
            elif feat_nm == "speed":
                sac_spd = dr / (dt + self.eps)
                feat_arr = sac_spd[transition_mask]
            else:
                raise NotImplemented(feat_nm)
            feats.append((feat_nm, feat_arr))

        return feats


class RegressionFeatures(StatsTransformer):
    """Regression Features Transformer.
    The transformer identifies saccades, and then selects regressions
    from them using user-defined set of rules.

    Args:
        rule: must be either 1) tuple of quadrants direction to classify
            regressions, 1st quadrant being upper-right square of plane and counting
            anti-clockwise or 2) tuple of angles in degrees (0 <= angle <= 360).
        deviation: if None, then `rule` is interpreted as quadrants. Otherwise,
            `rule` is interpreted as angles. If integer, then is a +-deviation for all angles.
            If tuple of integers, then must be of the same length as `rule`, each value being
            a corresponding deviation for each angle. Angle = 0 is positive x-axis direction,
            rotating anti-clockwise.
    """

    def __init__(
        self,
        rule: Tuple[int, ...],
        deviation: int | Tuple[int, ...] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.available_feats = ("length", "acceleration", "speed", "mask")
        self.rule = rule
        self.deviation = deviation

    @property
    def _fp(self) -> str:
        return "reg"

    def _check_params(self):
        if self.deviation is None:
            for q in self.rule:
                assert q in (1, 2, 3, 4), f"Wrong quadrant {q} in 'rule'."
        else:
            for a in self.rule:
                assert 0 <= a <= 360, f"Angles must be 0 <= angle <= 360, got {a}."
            if isinstance(self.deviation, int):
                assert 0 <= self.deviation <= 180, (
                    f"Deviation must be 0 <= deviation <= 180," f"got {self.deviation}."
                )
            elif isinstance(self.deviation, tuple):
                for d in self.deviation:
                    assert (
                        0 <= d <= 180
                    ), f"Deviation must be 0 <= deviation <= 180, got {d}."
            else:
                raise ValueError(
                    f"Wrong type for 'deviation': '{type(self.deviation)}'."
                )

        for feat in self.feature_names_in:
            assert self.x is not None, self._err_no_col(feat, "x")
            assert self.y is not None, self._err_no_col(feat, "y")
            if feat in ("speed", "acceleration"):
                assert self.t is not None
                self._err_no_col(feat, "t")

    def _calc_feats(
        self, X: pd.DataFrame, features: List[str], transition_mask: NDArray
    ) -> List[Tuple[str, pd.Series]]:
        feats = []

        dx: pd.Series = X[self.x].diff()
        dy: pd.Series = X[self.y].diff()
        sm = _select_regressions(dx, dy, self.rule, self.deviation)  # selection_mask
        dr = np.sqrt(dx**2 + dy**2)
        dt = (
            _calc_dt(X, self.duration, self.t)
            if any(map(lambda f: f != "length", features))
            else None
        )

        tm = transition_mask[sm]
        for feat_nm in features:
            if feat_nm == "length":
                sac_len = dr
                feat_arr = sac_len[sm][tm]
            elif feat_nm == "acceleration":
                sac_acc: pd.DataFrame = dr / (dt**2 + self.eps) * 1 / 2
                feat_arr = sac_acc[sm][tm]
            elif feat_nm == "speed":
                sac_spd = dr / (dt + self.eps)
                feat_arr = sac_spd[sm][tm]
            elif feat_nm == "mask":
                feat_arr = pd.Series(sm)
            else:
                raise NotImplemented(feat_nm)
            feats.append((feat_nm, feat_arr))

        return feats


class MicroSaccadeFeatures(StatsTransformer):
    """Micro Saccade Features.
    The transformer identities saccades, and then selects micro saccades
    from them using user-defined set of rules.

    Args:
        min_dispersion: minimum dispersion of fixation.
        max_speed: maximum speed between fixations.
    """

    def __init__(self, min_dispersion: float, max_speed: float, **kwargs):
        super().__init__(**kwargs)
        self.available_feats = ("length", "acceleration", "speed", "mask")
        self.min_dispersion = min_dispersion
        self.max_speed = max_speed

    @property
    def _fp(self) -> str:
        return "microsac"

    def _check_params(self):
        for feat in self.feature_names_in:
            assert self.x is not None, self._err_no_col(feat, "x")
            assert self.y is not None, self._err_no_col(feat, "y")
            assert self.dispersion is not None, self._err_no_col(feat, "dispersion")
            if feat in ("speed", "acceleration"):
                assert self.t is not None
                self._err_no_col(feat, "t")

    def _calc_feats(
        self, X: pd.DataFrame, features: List[str], transition_mask: NDArray
    ) -> List[Tuple[str, pd.Series]]:
        feats = []

        dx: pd.Series = X[self.x].diff()
        dy: pd.Series = X[self.y].diff()
        dr = np.sqrt(dx**2 + dy**2)

        # selection_mask
        sm = (dr < self.max_speed) & (X[self.dispersion] > self.min_dispersion)

        dt = (
            _calc_dt(X, self.duration, self.t)
            if any(map(lambda f: f != "length", features))
            else None
        )

        tm = transition_mask[sm]
        for feat_nm in features:
            if feat_nm == "length":
                sac_len = dr
                feat_arr = sac_len[sm][tm]
            elif feat_nm == "acceleration":
                sac_acc: pd.DataFrame = dr / (dt**2 + self.eps) * 1 / 2
                feat_arr = sac_acc[sm][tm]
            elif feat_nm == "speed":
                sac_spd = dr / (dt + self.eps)
                feat_arr = sac_spd[sm][tm]
            elif feat_nm == "mask":
                feat_arr = pd.Series(sm)
            else:
                raise NotImplemented(feat_nm)
            feats.append((feat_nm, feat_arr))

        return feats


class FixationFeatures(StatsTransformer):
    """Fixation Features Transformer.
    The transformer uses input fixations to extract features.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.available_feats = ("duration", "vad")

    @property
    def _fp(self) -> str:
        return "fix"

    def _check_params(self):
        for feat in self.feature_names_in:
            if feat == "duration":
                assert self.duration is not None, self._err_no_col(feat, "duration")
            elif feat == "vad":
                assert self.dispersion is not None, self._err_no_col(feat, "dispersion")

    def _calc_feats(
        self, X: pd.DataFrame, features: List[str], transition_mask: NDArray
    ) -> List[Tuple[str, pd.Series]]:
        feats = []

        for feat_nm in features:
            if feat_nm == "duration":
                feat_arr = X[self.duration][transition_mask]
            elif feat_nm == "vad":
                feat_arr = X[self.dispersion][transition_mask]
            else:
                raise NotImplemented(feat_nm)
            feats.append((feat_nm, feat_arr))

        return feats
